#include "preprocessor/mesh_metadata.hpp"
#include "preprocessor/input_parameters.hpp"
#include "util.hpp"

#include <deque>
#include <unordered_set>

void write_distributed_edge_metadata(const std::string& file_name,
                                     const InputParameters<>& input,
                                     const MeshMetaData& mesh_meta,
                                     const std::vector<std::vector<MeshMetaData>>& submeshes) {
    std::size_t num_loc = submeshes.size();

    std::unordered_set<std::pair<uint, uint>> faces;
    // assemble faces shared across elements
    for (auto& elt : mesh_meta.elements) {
        for (uint i = 0; i < elt.second.neighbor_ID.size(); ++i) {
            if (is_internal(elt.second.boundary_type[i])) {
                uint neigh{elt.second.neighbor_ID[i]};
                std::pair<uint, uint> edge_name{std::min(elt.first, neigh), std::max(elt.first, neigh)};
                faces.insert(edge_name);
            }
        }
    }
    std::unordered_map<uint, uint> elt2partition;
    for (uint loc_id = 0; loc_id < submeshes.size(); ++loc_id) {
        for (uint sbmsh_id = 0; sbmsh_id < submeshes[loc_id].size(); ++sbmsh_id) {
            for (auto& elt : submeshes[loc_id][sbmsh_id].elements) {
                elt2partition.insert(std::make_pair(elt.first, loc_id + num_loc * sbmsh_id));
                for (uint i = 0; i < elt.second.neighbor_ID.size(); ++i) {
                    if (is_internal(elt.second.boundary_type[i])) {
                        uint neigh{elt.second.neighbor_ID[i]};
                        std::pair<uint, uint> edge_name{std::min(elt.first, neigh), std::max(elt.first, neigh)};
                        faces.erase(edge_name);
                    }
                }
            }
        }
    }
    // Assemble lists of distributed interface meta data
    std::unordered_map<std::pair<uint, uint>, std::deque<DBPairMetaData>> shared_faces;
    for (auto& f : faces) {
        uint eltA = f.first;
        uint rnkA = elt2partition.at(eltA);

        uint eltB = f.second;
        uint rnkB = elt2partition.at(eltB);

        uint face_id_A{DEFAULT_ID};
        const ElementMetaData& eltA_meta = mesh_meta.elements.at(eltA);
        for (uint fid = 0; fid < eltA_meta.neighbor_ID.size(); ++fid) {
            if (eltB == eltA_meta.neighbor_ID[fid]) {
                face_id_A = fid;
            }
        }
        assert(face_id_A != DEFAULT_ID);

        uint face_id_B{DEFAULT_ID};
        const ElementMetaData& eltB_meta = mesh_meta.elements.at(eltB);
        for (uint fid = 0; fid < eltB_meta.neighbor_ID.size(); ++fid) {
            if (eltA == eltB_meta.neighbor_ID[fid]) {
                face_id_B = fid;
            }
        }
        assert(face_id_B != DEFAULT_ID);

        std::pair<uint, uint> rnk_pair{std::min(rnkA, rnkB), std::max(rnkA, rnkB)};

        DBPairMetaData dist_int;
        dist_int.p = input.polynomial_order;

        if (rnkA > rnkB) {
            dist_int.elements  = {eltB, eltA};
            dist_int.bound_ids = {face_id_B, face_id_A};
        } else {
            dist_int.elements  = {eltA, eltB};
            dist_int.bound_ids = {face_id_A, face_id_B};
        }

        shared_faces[rnk_pair].push_back(std::move(dist_int));
    }

    for (uint loc_id = 0; loc_id < submeshes.size(); ++loc_id) {
        for (uint sbmsh_id = 0; sbmsh_id < submeshes[loc_id].size(); ++sbmsh_id) {
            std::string distributed_meta_filename = file_name;
            distributed_meta_filename =
                distributed_meta_filename.substr(0, distributed_meta_filename.find_last_of("."));
            std::ofstream file(distributed_meta_filename + '_' + std::to_string(loc_id) + '_' +
                               std::to_string(sbmsh_id) + ".dbmd");
            for (auto& sf : shared_faces) {
                uint locA   = sf.first.first % num_loc;
                uint sbmshA = sf.first.first / num_loc;

                uint locB   = sf.first.second % num_loc;
                uint sbmshB = sf.first.second / num_loc;
                if ((locA == loc_id && sbmshA == sbmsh_id) || (locB == loc_id && sbmshB == sbmsh_id)) {
                    file << locA << " " << sbmshA << " " << locB << " " << sbmshB << " " << sf.second.size() << '\n';
                    for (auto& dist_int : sf.second) {
                        file << dist_int << '\n';
                    }
                }
            }
        }
    }
}
